<?php

namespace Drupal\office_hours\Plugin\Field\FieldFormatter;

use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\office_hours\Plugin\Field\FieldType\OfficeHoursItem;

/**
 * Plugin implementation of the formatter.
 *
 * @FieldFormatter(
 *   id = "office_hours_table",
 *   label = @Translation("Table"),
 *   field_types = {
 *     "office_hours",
 *   }
 * )
 */
class OfficeHoursFormatterTable extends OfficeHoursFormatterDefault {

  /**
   * {@inheritdoc}
   */
  public function settingsSummary() {
    $summary = parent::settingsSummary();
    $summary[] = $this->t('Display Office hours in a table.');
    return $summary;
  }

  /**
   * {@inheritdoc}
   */
  public function viewElements(FieldItemListInterface $items, $langcode) {
    $elements = parent::viewElements($items, $langcode);

    // If no data is filled for this entity, do not show the formatter.
    if ($items->isEmpty()) {
      return $elements;
    }

    // Fetch the correct element from the already formatted data.
    $hours_formatter = NULL;
    foreach ($elements as $key => $element) {
      switch ($element['#theme'] ?? '') {
        case 'office_hours':
        case 'office_hours_table':
          // Fetch the Office Hours formatter.
          $hours_formatter = &$elements[$key];
          // Leave loop to preserve $key.
          break 2;
      }
    }

    if (!$hours_formatter) {
      return $elements;
    }

    $settings = $this->getSettings();
    $field_definition = $items->getFieldDefinition();

    // Determine the required columns.
    $row_elements['label'] = $settings['day_format'] != 'none';
    $row_elements['slots'] = TRUE;
    $row_elements['comments'] = $this->getFieldSetting('comment');

    $table_weight = $hours_formatter['#weight'];
    $table_index = $key;
    // Build the table header for each column.
    $header = $this->buildTableHeader($row_elements);
    // Build a table element with 0 rows.
    $table_caption = '';
    $table_class = '';
    $table = $this->buildTable($table_caption, $field_definition, $header, $table_rows = [], $table_class);

    // Overwrite parent.
    $hours_formatter['#theme'] = 'office_hours_table';
    $hours_formatter['#table'] = $table;

    // Build/add the table rows.
    $office_hours = $hours_formatter['#office_hours'];
    foreach ($office_hours as $delta => $info) {

      if (TRUE) {
        // Use FALSE/TRUE to quickly test/toggle between 2 formats in #3355646.
        $hours_formatter['#table']['#rows'][$delta] = $this->buildTableRow($info, $row_elements);
      }

    }

    return $elements;
  }

  public function buildTable(?string $table_caption, FieldDefinitionInterface $field_definition, array $table_header, array $table_rows, ?string $table_class): array {
    return [
      '#theme' => 'table',
      '#parent' => $field_definition,
      '#caption' => $table_caption,
      '#header' => $table_header,
      // '#empty' => $this->t('This location has no opening hours.'),
      '#rows' => $table_rows,
      '#attributes' => [
        'class' => ['office-hours__table', $table_class],
      ],
      '#attached' => [
        'library' => [
          'office_hours/office_hours_formatter',
        ],
      ],
    ];
  }

  protected function buildTableHeader(array $row_elements): array {
    // Add a label/header/title for accessibility (a11y) screen readers.
    // Superfluous comments are removed. @see #3110755 for examples.
    $element = [];

    $labels = OfficeHoursItem::getPropertyLabels('data', $this->getFieldSettings() + ['slots' => TRUE]);
    if ($row_elements['label']) {
      $element['label'] = [
        'data' => $labels['day']['data'],
        'class' => 'visually-hidden',
      ];
    }
    if ($row_elements['slots']) {
      $element['slots'] = [
        'data' => $labels['slots']['data'],
        'class' => 'visually-hidden',
      ];
    }
    if ($row_elements['comments']) {
      $element['comments'] = [
        'data' => $labels['comment']['data'],
        'class' => 'visually-hidden',
      ];
    }
    return $element;
  }

  protected function buildTableRow(array $info, array $row_elements): array {
    // Add a label/header/title for accessibility (a11y) screen readers.
    // Superfluous comments are removed. @see #3110755 for examples.
    $element = [
      'no_striping' => TRUE,
      'class' => ['office-hours__item'],
    ];

    // N.B. 'Show current day' may return nothing in getRows(),
    // while other days are filled.
    if ($info['is_current']) {
      $element['class'][] = 'office-hours__item-current';
    }

    $element['data'] = [];
    if ($row_elements['label']) {
      $element['data']['label'] = [
        'data' => ['#markup' => $info['label']],
        // Switch 'Day' between <th> and <tr>.
        'header' => !$row_elements['comments'],
      ];

      $item = $info['items'][0] ?? NULL;
      switch (TRUE) {
        case is_null($item):
          $element['data']['label']['class'] = ['office-hours__item-label'];
          break;

        case $item->isExceptionHeader():
        case $item->isSeasonHeader():
          $element['data']['label']['class'] = ['office-hours__exceptions-label'];
          break;

        default:
          $element['data']['label']['class'] = ['office-hours__item-label'];
          break;
      }
    }

    if ($row_elements['slots']) {
      $element['data']['slots'] = [
        'data' => ['#markup' => $info['formatted_slots']],
        'class' => ['office-hours__item-slots'],
      ];
    }

    if ($row_elements['comments']) {
      $element['data']['comments'] = [
        'data' => ['#markup' => $info['comments']],
        'class' => ['office-hours__item-comments'],
      ];
    }

    return $element;
  }

}
